#include "logger.h"

void DpcRoutine(IN PKDPC Dpc, IN PVOID DeferredContext, IN PVOID SystemArgument1, IN PVOID SystemArgument2);
void LoggerThread(__in PVOID  StartContext);

MultiLogger multilog;
ULONG g_DoLog;
HANDLE g_ThreadHandle;

NTSTATUS logger_init(ULONG max)
{
	CCHAR Num;

	multilog.nCore = KeNumberProcessors;
	
	multilog.Logs = ExAllocatePool(NonPagedPool, multilog.nCore * sizeof(LogsQueue));
	if (!multilog.Logs){
		return STATUS_UNSUCCESSFUL;
	}

	for (Num = 0; Num < multilog.nCore; Num++){
		multilog.Logs[Num].length = max;
		multilog.Logs[Num].tail = 0;
		multilog.Logs[Num].head = 0;
		multilog.Logs[Num].queue = ExAllocatePool(NonPagedPool, max * sizeof(LogData));
	}
	
	return STATUS_SUCCESS;
}

NTSTATUS logger_start()
{
	g_DoLog = 1;
	
	PsCreateSystemThread(&g_ThreadHandle, 0, NULL, NULL, NULL, (PKSTART_ROUTINE) LoggerThread, NULL);
	
	return STATUS_SUCCESS;
}

void logger_stop(void)
{
	InterlockedExchange(&g_DoLog, 0);
}

void logger_destroy()
{
	int i;
	
	for (i = 0; i < multilog.nCore; i++){
		if (MmIsAddressValid(multilog.Logs[i].queue)){
			ExFreePoolWithTag(multilog.Logs[i].queue, 0);
		}
	}
	
	if (MmIsAddressValid(multilog.Logs)){
		ExFreePoolWithTag(multilog.Logs, 0);
	}
}

int logger_enqueue(int Cpu, ULONG cs, ULONG eip, ULONG fault_address, ULONG error_code)
{
	LogsQueue *log_queue;
	int bStatus = 0;
	int t;
	
	if (Cpu > multilog.nCore) return 0;
	if (multilog.Logs[Cpu].queue == 0) return 0;
	
	log_queue = &multilog.Logs[Cpu];
	
	t = (log_queue->tail + 1) % log_queue->length;
	
	if (t != log_queue->head){
		log_queue->queue[t].CS = cs;
		log_queue->queue[t].EIP = eip;
		log_queue->queue[t].FaultAddress = fault_address;
		log_queue->queue[t].ErrorCode = error_code;
		
		log_queue->tail = t;
		bStatus = 1;
	}
	
	return bStatus;
}

int logger_dequeue(int Cpu, LogData *log)
{
	LogsQueue *log_queue;
	int bStatus = 0;
	int t;
	
	if (Cpu > multilog.nCore) return 0;
	if (multilog.Logs[Cpu].queue == 0) return 0;
	
	RtlZeroMemory(log, sizeof(LogData));
	log_queue = &multilog.Logs[Cpu];

	t = (log_queue->head + 1) % log_queue->length;
	
	if (t != log_queue->tail){
		log->CS = log_queue->queue[t].CS;
		log->EIP = log_queue->queue[t].EIP;
		log->FaultAddress = log_queue->queue[t].FaultAddress;
		log->ErrorCode = log_queue->queue[t].ErrorCode;
		
		log_queue->head = t;
		bStatus = 1;
	}
	
	return bStatus;
}

void LoggerThread(__in PVOID  StartContext)
{
	LARGE_INTEGER large_int;
	LogData log;
	KTIMER timer;
	int i;
	
	large_int.QuadPart = - 10 * 1000 * 1000 * 2; /* this is for one second */

	while(InterlockedExchange(&g_DoLog, 1)){
		KeInitializeTimer(&timer);
		KeSetTimer(&timer, large_int, NULL);
		KeWaitForSingleObject(&timer, Executive, KernelMode, FALSE, NULL);	
		
	
		DbgPrint("[LOG].START\n");
		while(logger_dequeue(KeGetCurrentProcessorNumber(), &log)){
			if (log.CS && log.EIP){
				DbgPrint("[LOG]\tcs:%04X\teip:%08X\tfault:%08X\terror:%08X\n", log.CS, log.EIP, log.FaultAddress, log.ErrorCode);
			}
		}
		DbgPrint("[LOG].END\n");
		
		KeCancelTimer(&timer);
	}
	
	PsTerminateSystemThread(0);
}